(* $Id: godi_match.mli 568 2005-07-22 15:10:00Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann. 
 * 
 * This file is part of godi-tools. 
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

(* Variants of regular expressions *)

(**********************************************************************)
(* GLOBBING, FNMATCH                                                  *)
(**********************************************************************)

type glob_expr = 
     private
       Glob of glob_expr_atom list
  (* A parsed glob expression. It is exported as private type, so 
   * you cannot create such expressions directly, only by [parse_glob_expr].
   *)

and glob_expr_atom = 
    private 
    G_literal of string
      (* Matches the string literally. The string must not be empty. 
       * The backslash is not an escape character, but matches the
       * backslash character. Two adjecent G_literal atoms are
       * not allowed.
       *)
  | G_star
      (* The "*" operator *)
  | G_qmark
      (* The "?" operator *)
  | G_bracket of (bool * set)
      (* The "[...]" operator. The set argument describes the characters
       * that are matched. The bool argument is true when the expression
       * is negated ([^...]). 
       *)
  | G_brace of glob_expr list
      (* The "{e1,e2,...}" operator. *)

and set =
    (int * int) list
  (* Sets must not be empty. For every pair (x,y): x<=y *)


exception Bad_glob_expr of string
  (* An error in the glob expression; the argument is the bad expression *)

exception Bad_regular_expr of string
  (* An error in the regexp expression; the argument is the bad expression *)

exception Unsupported_expr of string
  (* The notations [:class:], [.symbol.], [=eqclass=] inside [...] are
   * not supported by this implementation. If they are found, this exception
   * will be raised, and the argument is the whole glob or regexp expression
   *)

val parse_glob_expr : 
           ?enable_star:bool ->     (* Recognize "*" *)
           ?enable_qmark:bool ->    (* Recognize "?" *)
           ?enable_brackets:bool -> (* Recognize "[set]" *)
           ?enable_braces:bool ->   (* Recognize "{alt,...}" *)
           ?enable_escape:bool ->   (* Recognize backslash as escape char *)
           string ->
	     glob_expr
  (* Parses the glob expression. By default, all syntax features are enabled.
   * May raise Bad_glob_expr or Unsupported_expr.
   *
   * The glob expressions are POSIX-compliant with the extension of 
   * brace expressions and the omission of internationalized bracket
   * expressions:
   * - "*": Matches a sequence of zero or more arbitrary characters
   * - "?": Matches one arbitrary character
   * - "[abc]": Matches one of the mentioned characters
   * - "[a-z]": Matches one of the characters of the range (POSIX locale!)
   * - "[!expr]" or "[^expr]": Negates the bracket expression
   * - "{expr,expr,...}": Generates a string for each of the alternatives.
   *   A brace expression is even recognized if there is no comma, or even
   *   no contents (i.e. "{expr}" and "{}"). The elements of brace expressions
   *   may be again glob expressions; nested brace expressions are allowed.
   * - If enabled, the backslash character is the escape character. Within
   *   bracket expressions, the backslash character does never escape.
   * - Not supported: Collating symbols "[.a.]", equivalence classes
   *   "[=a=]", and character classes "[:name:]". If thery are found, the
   *   exception Unsupported_expr will be raised.
   *)

val print_glob_expr :
           glob_expr ->
	     string
  (* Prints the glob expression as string. Meta characters are always
   * escaped by a backslash. Meta characters are:
   *   "*", "?", "[", "]", "{", "}", ",", and "\\" 
   *)

val expand_braces : glob_expr -> glob_expr list
  (* Resolve "{...}" sub expressions by expanding them. The returned
   * list of glob_expr no longer contains any brace expressions.
   *)

val regexp_of_glob_expr : 
        ?protect_period:bool ->   (* Protect leading dots; default: true *)
	?protect_slash:bool ->    (* Protect slashes; default: true *)
	glob_expr -> 
	  string
  (* Transforms the glob_expr into a regexp. The regexp is anchored
   * (^ at beginning, $ at end), and compatible with the Str module.
   *
   * The glob_expr must not contain brace expressions (i.e. call
   * expand_braces first). The function fails if it encounters such an
   * expression.
   *
   * ~protect_period: If true, a leading period cannot be not matched by
   *    "*", "?", "[...]", but only by a literal ".", or by a "{...}"
   *    expression. A leading period is:
   *      - a "." at the beginning of the string to be matched against
   *      - if also ~protect_slash: a "." after a "/"
   *
   * ~protect_slash: If true, a slash cannot be matched by "*", "?", "[...]",
   *    but only by a literal "/", or by a "{...}" expression.
   *
   * Both options are enabled by default.
   *)

val split_glob_expr : glob_expr -> glob_expr list
  (* Splits the glob expression into filename components separated by
   * literal '/' characters. For example, for the glob expression
   * "a*b/c/d?", the list ["a*b"; "c"; "d?"] is returned. 
   * 
   * If the first component begins with a slash, the slash is not removed
   * from the first returned list element, e.g. for "/ab/c*", the list
   * [ "/ab"; "c*" ] is computed. Use check_rooted_glob_expr to test this
   * case.
   *
   * Several adjacent slashes are handled like a single slash. E.g.
   * for "a//b", the list ["a"; "b"] is returned.
   *
   * If the last component ends with a slash, it is not removed from the
   * returned list element, e.g. for "a/b/", the list [ "a"; "b/" ] is
   * returned. Use check_directory_glob_expr to test this case.
   *
   * The glob expression passed to this function must not contain brace
   * expressions.
   *)

val check_rooted_glob_expr : glob_expr -> glob_expr option
  (* If the glob expression matches the root directory (i.e. the expression
   * begins with a literal '/'), the function returns Some expr', where
   * expr' matches the path relative to the root directory (i.e. the
   * expression without the '/' at the beginning).
   *
   * Otherwise, None is returned.
   *
   * Example: For "/a/b*", the expression "a/b*" is returned.
   *
   * Special case: for "/", the expression "" (only matching the empty string)
   * is returned.
   *
   * The glob expression passed to this function must not contain brace
   * expressions.
   *)

val check_directory_glob_expr : glob_expr -> glob_expr option
  (* If the last component of the glob expression matches only directories
   * because it ends with a literal '/' character, the value Some expr' is
   * returned where expr' matches the same path without the trailing '/'.
   *
   * Otherwise, None is returned.
   *
   * Example: For "a/b*/", the expression "a/b*" is returned.
   *
   * Special case: for "/", the expression "" (only matching the empty string)
   * is returned.
   *
   * The glob expression passed to this function must not contain brace
   * expressions.
   *)


type fsys =
    (* File system operations for [glob] *)
    { fsys_read_dir : string -> string list;
        (* Returns the file names contained in the directory, without
	 * path. The names "." and ".." should be returned. It is acceptable
	 * to return the empty list for an unreadable directory.
	 *)
      fsys_file_is_dir : string -> bool;
        (* Whether the file name is valid and a directory, or a symlink to
	 * a directory.
	 *)
      fsys_file_exists : string -> bool;
        (* Whether the file name is valid and refers to an existing file,
	 * or to a symlink pointing to an existing file.
	 *)
    }
;;


type glob_mode = [ `Existing_paths
		 | `All_paths
		 | `All_words
		 ]


val glob :
      ?base_dir:string ->       (* default: current directory *)
      ?protect_period:bool ->   (* default: true *)
      ?fsys:fsys ->             (* default: access real file system *)
      ?mode:glob_mode ->        (* default: `Existing_paths *)
      glob_expr ->
	string list
  (* Forms a set of filenames as described below, and matches this set
   * against glob_expr. In this match, "/" is always protected, and must
   * literally occur in order to be a candidate for matching. The leading
   * period is protected if ~protect_period.
   *
   * If the glob_expr begins with a literal "/", the set of filenames is
   * anchored at the root directory; otherwise the set is anchored at
   * the current directory or, if ~base_dir is passed, at this directory.
   * 
   * Initially, the set contains all files of the anchor
   * directory (for the root directory, a "/" is prepended).
   *
   * After that, the set is extended by adding the paths of
   * subdirectories relative to the anchor directory. Note that the
   * constructed set is always infinite, because "." and ".." are not
   * handled specially, and are also regarded as "subdirectories". However,
   * after applying the matching criterion, the returned list is always
   * finite.
   *
   * Note that the anchor directory itself is not part of the generated
   * set. For example, for the expression "/*" the root directory "/" is
   * not returned. As an exception of this rule, for the glob expression
   * "/" the file "/" is returned.
   *
   * Brace expressions are handled by expanding them first, even before
   * filename generation starts.
   *
   * If no files match, the empty list is returned.
   *
   * By default, only existing paths are returned. By passing a
   * different ~mode, this can be changed:
   * - `All_paths: It is allowed that non-existing paths
   *    are returned when the paths do not contain *, ?, or [
   *    metacharacters after the brace expansion. Path expressions
   *    with these metacharacters are still checked for existence.
   * - `All_words: When an expression does not refer to existing
   *    paths, it is returned as such, leaving the metacharacters *, ?, [
   *    unexpanded (i.e., what the Bourne shell does). Note that
   *    either all metacharacters are resolved, or none, but not
   *    a subset of them.
   *)


(**********************************************************************)
(* STRING SUBSTITUTIONS                                               *)
(**********************************************************************)

(* A substitution matches a string against a pattern, and computes
 * a replacement text where parts of the pattern can be referenced.
 *)

type subst_expr

exception No_percent_symbol
  (* See parse_percent_subst_expr *)

exception Bad_subst_expr
  (* Raised when one of the parsers fails *)

val parse_percent_subst_expr : 
      ?indicate_no_percent:bool ->   (* default: false *)
      ?enable_escape:bool ->         (* default: true *)
      (string * string) ->
	subst_expr
  (* "percent substitution": The lhs and the rhs may contain the '%'
   * symbol. On the lhs '%' matches any number of characters. On the
   * rhs '%' is replaced by the matched characters.
   *
   * Example: A%B=B%A changes AtextB into BtextA.
   *
   * On the lhs, there must be only one '%' symbol. The lhs is
   * anchored at the begining and at the end.
   *
   * ~indicate_no_percent: If true, the exception No_percent_symbol
   * is raised if the lhs does not contain a '%' symbol.
   *
   * ~enable_escape: If true, the backslash is the escape character.
   *)

val parse_suffix_subst_expr :
      ?enable_escape:bool ->        (* default: true *)
      (string * string) ->
	subst_expr
  (* Suffix substitution: If the lhs matches the suffix of the string,
   * it is replaced by the rhs.
   *
   * Example: .c=.o changes file.c into file.o
   *
   * ~enable_escape: If true, the backslash is the escape character.
   *)

val parse_substring_subst_expr :
      ?enable_anchors:bool ->       (* default: true *)
      ?enable_escape:bool ->        (* default: true *)
      (string * string) ->
	subst_expr
  (* Substring substitution: The lhs is the substring to match. The
   * rhs is the replacement text, where the '&' character is replaced
   * by the matched substring.
   *
   * Example: abc=x&y changes uabcv into uxabcyv
   *
   * ~enable_anchors: If true, the character '^' at the beginning of the
   *   lhs anchors the pattern at the beginning, and the character '$'
   *   at the end of the rhs anchors the pattern at the end.
   *
   * ~enable_escape: If true, the backslash is the escape character.
   *)

val create_subst_expr : Str.regexp -> string -> subst_expr
  (* Creates a substitution from a regular expression and a replacement
   * text. The latter can include references to groups "\N" where N
   * is the number of the group (N from 1 to 9), and "\0" meaning the
   * whole match. 
   *)

val substitute :
      ?globally:bool ->            (* default: false *)
      subst_expr ->
      string ->
	string
  (* Substitutes matching patterns of the input string according
   * to the subst_expr.
   *
   * By default, only the first match is substituted. If ~globally=true,
   * all matches are substituted.
   *)


(**********************************************************************)
(* EXTENDED REGULAR EXPRESSIONS                                       *)
(**********************************************************************)

val extended_regexp :
      string -> Str.regexp
  (* Parses an extended regular expression, see regex(7).
   * May raise Bad_regular_expr or Unsupported_expr.
   *
   * Limitation: The syntax "expr{n,m}" is recognized, but there is no
   * way to translate it to Str-compatible regexps. Unsupported_expr is
   * raised in this case.
   *)

(**********************************************************************)
(* UNQUOTING SHELL LISTS                                              *)
(**********************************************************************)

val parse_shell_words : string -> string list
  (* Parses a string as shell-quoted list of words. Backslash, single
   * and double quotes are supported. Other shell metacharacters are
   * simply parsed literally.
   *)

val quote_shell_word : string -> string
  (* Quotes a single word *)

val quote_shell_words : string list -> string
  (* Quotes as a list of words *)
